import altair as alt
import pandas as pd
import numpy as np


def create_hist_dataframe(data=None, *, x=None, y=None):
  # create data if x and y are pandas series
  if data is None:
    data = pd.DataFrame({})
    # case that x series is provided 
    if isinstance(x, pd.Series):
      data['x'] = x
      x = 'x'

    # case that y series is provided 
    if isinstance(y, pd.Series):
      data['y'] = y
      y = 'y'
  
  return data,x,y

def hist(data=None,x=None,y=None, width=200,height=50,filters=None,color=None,fill="steelblue",xAxis = alt.Axis(),yAxis=alt.Axis(),interactive=False):
  if filters is None:
    filters = []
    
  # ensures that data is the data and x and y are column names
  data,x,y = create_hist_dataframe(data=data,x=x,y=y) 
  chart = None

  layers = {"fg":None,"bg":None}

  if x is not None and y is None:
    layers['fg']= alt.Chart(data).mark_bar(color=fill).encode(
            alt.X(f'{x}:Q', bin=True, axis=xAxis),alt.Y('count()',axis=yAxis)
              ) 
    layers['bg'] = alt.Chart(data).mark_bar(color='lightgray').encode(
        alt.X(f'{x}:Q', bin=True, axis=xAxis),alt.Y('count()',axis=yAxis)
      )
    
    if interactive:

      brush = alt.selection_interval(encodings=['x'],resolve="union",name='brush')
      
      if type(interactive) == type(alt.selection_interval()):
        brush = interactive     
      
      layers['fg'] =  layers['fg'].add_selection(brush)
      filters.append(brush)


  elif x is  None and y is not None:
    layers['fg']= alt.Chart(data).mark_bar(color=fill).encode(
            alt.Y(f'{y}:Q', bin=True, axis=yAxis),alt.X('count()',axis=xAxis)
              ) 
    layers['bg'] = alt.Chart(data).mark_bar(color='lightgray').encode(
        alt.Y(f'{y}:Q', bin=True, axis=yAxis),alt.X('count()',axis=xAxis)
      )
    
    if interactive:

      brush = alt.selection_interval(encodings=['y'],resolve="union",name='brush')
      
      if type(interactive) == type(alt.selection_interval()):
        brush = interactive     

      
      layers['fg'] =  layers['fg'].add_selection(brush)
      filters.append(brush)
  

  if filters:
     for filter in filters:
        layers['fg'] = layers['fg'].transform_filter(filter)

  chart = layers['bg'] + layers['fg'] 
  
  return chart.properties(
          width=width,
          height=height
      )



def violin_plots(data=None,y=None,groupby=None, yAxis=None,xAxis=alt.Axis(labels=False, values=[0],grid=False, ticks=True),interactive=False,filters=None):
  if filters is None:
    filters = []

  facet_vars = [None]
  if groupby:
    facet_vars=np.unique(data[groupby])

  brush = alt.selection_interval(
    name='brush',
    encodings=['y'],
    empty="all"
  )

  charts =[]

  for index,variable in enumerate(facet_vars):
    # filter to unique value
    base = alt.Chart(data=data)

    # filter to only one variable
    if variable is not None:
      base=base.transform_filter(
          alt.FieldEqualPredicate(field=groupby, equal=variable)
      )

    if yAxis is None:
      if index == 0:
        yAxis = alt.Axis(grid=False, ticks=True)
    else:
      if index != 0:
        yAxis = None
          
    
    
    

    layers = {'fg':None,'bg':None}
    # for each value in origin, 
      # create layered plot
    # concat plots together

    layers['bg'] = base.mark_area(color="lightgray",
                                  ).transform_density(
        y,
        as_=[y, 'density'],
        extent=[5, 55],
    ).transform_stack(
        stack= "density",
        groupby= [y],
      as_= ["x", "x2"],
      offset= "center"
    ).encode(
                    
        y=alt.Y(f'{y}:Q',axis=yAxis),
        x=alt.X(
            field='x',
            impute=None,
            title=None,
            type ="quantitative",
            axis=xAxis,
        ),
            x2=alt.X2(field = "x2")

    )


    layers['fg'] = base
    if interactive:
      if type(interactive) == type(alt.selection_interval()):
        brush = interactive     
      layers['bg'] =  layers['bg'].add_selection(brush)
      filters.append(brush)
  

    if filters:
      for filter in filters:
        layers['fg'] = layers['fg'].transform_filter(filter)
    
    
    layers['fg'] =layers['fg'].transform_density(
        y,
        as_=[y, 'density'],
        extent=[5, 55],
    ).transform_stack(
        stack= "density",
        groupby= [y],
      as_= ["x", "x2"],
      offset= "center"
    )

    layers['fg'] = layers['fg'].mark_area(orient='horizontal', align="center",
                                  ).encode(
        y=alt.Y(f'{y}:Q',axis=yAxis),
        x=alt.X(
            field='x',
            impute=None,
            title=None,
            type ="quantitative",
                        axis=xAxis,

        ),
        x2=alt.X2(field = "x2")
    )
                  
    chrt = layers['bg'] + layers['fg']
    charts.append(chrt.properties(width=100,title = alt.TitleParams(text = variable )))

  final_chart = None
  for chart in charts:
    if final_chart is None:
      final_chart = chart
    else:
      final_chart = alt.hconcat(final_chart,chart,spacing=0)
  return final_chart


def countplot(data=None,x=None,xAxis=alt.Axis(),yAxis=alt.Axis(), interactive=False, filters=[],width=250,height=150, max_bars = None):
  if data is None:
    if x is None:
      raise ValueError('[countplot] no data or data series provided.')
    data = pd.DataFrame({})
    if isinstance(x, pd.Series):
      data['x'] = x
      x = 'x'

  layers = {"fg":alt.Chart(data).mark_bar(color="steelblue")
    .encode(
      alt.X(f'{x}:N',axis=xAxis), # remove the sort as that will keep it consistent with the background
            alt.Y(f'count({x}):Q',axis=yAxis)
  ),"bg":alt.Chart(data).mark_bar(color="lightgray")
    .encode(
      alt.X(f'{x}:N',sort='-y',axis=xAxis), 
      alt.Y(f'count({x}):Q',axis=yAxis)
  )}

  if interactive:
      x_brush = alt.selection_multi(name='brush',encodings=['x'],resolve='union')
      if type(interactive) ==type(alt.selection_interval()):
        x_brush = interactive     
      
      layers['bg'] =  layers['bg'].add_selection(x_brush)
      filters.append(x_brush)
   
  if filters:
     for filter in filters:
        layers['fg'] = layers['fg'].transform_filter(filter)

  chart = layers['bg'] + layers['fg'] 
  return chart.properties(
          width=width,
          height=height
      )


import altair as alt
from .distributional import hist
from .relational import scatterplot

def jointplot(data=None,x=None, y=None):
  x_brush = alt.selection_interval(encodings=['x'],resolve="intersect",name="brush")
  y_brush = alt.selection_interval(encodings=['y'],resolve="intersect",name="brush")

  top = hist(data=data,x=x,width=200,height=50,xAxis=None,yAxis=None,interactive=x_brush)
  right = hist(data=data,y=y,width=50,height=200,xAxis=None,yAxis=None,interactive=y_brush)

  mid = scatterplot(data,x=x,y=y,filters=[x_brush,y_brush],interactive=False)


  # question is there a way to 
  return alt.vconcat(top, alt.hconcat(mid,right,spacing=-10), spacing=-10)



import altair as alt
from .distributional import hist, countplot
from pandas.api.types import is_numeric_dtype
import pandas as pd
import numpy as np

def data_table(data=None,vars=None,row_limit=25):
  if vars is None:
    vars = data.columns
  
  # sort so they're in line with the table marks
  if type(vars) is list:
    vars.sort()
  
  brush = alt.selection_interval(name='brush', resolve="intersect")
  picker = alt.selection_single(name='picker',encodings=['x'],resolve='union')
  charts = []

  column_width = 90

  chart_padding = 15
  chart_width = column_width-chart_padding

  for index, variable in enumerate(vars):
    if index == 0:
         spacer = alt.Chart(data=data).mark_circle(size=2,opacity=0).encode(
             alt.X(variable,axis=None),
             alt.Y(variable,axis=None)
         ).properties(width=chart_width-50, height=50)
         charts.append(spacer)
         
    chart_index = index + 1 # for spacer
    if(is_numeric_dtype(data[variable])):

      charts.append(hist(data = data,x = variable,width=chart_width,interactive=brush,xAxis=None,yAxis=None,filters=[picker]))

      brush_name = brush.name
      signal_index = f'{brush_name}.{variable}'
      max=np.max(data[variable])
      min=np.min(data[variable])
      bounds = alt.hconcat(alt.Chart(data=pd.DataFrame({'min':['']}))
            .transform_calculate(calculate=f"{signal_index} ? +format({signal_index}[0],'.2f'):+format({min},'.2f')", as_="min")
            .mark_text()
            .encode(alt.Text(field='min')),
          alt.Chart(data=pd.DataFrame({'max':['']})).transform_calculate(calculate=f"{signal_index} ? +format({signal_index}[1],'.2f'):+format({max},'.2f')", as_="max")
            .mark_text()
            .encode(alt.Text(field='max'))
          
          )
      
      
      charts[chart_index] = alt.vconcat(charts[chart_index],bounds, spacing=0)
    else:
      # append blank till count plot are ready
      spacer = countplot(data=data,x=variable,interactive=picker,width=column_width,height=50,filters=[brush],xAxis=None,yAxis=None)
      signal_index = False
      bounds = alt.hconcat(alt.Chart(data=pd.DataFrame({'min':['_']}))
            .mark_text()
            .encode(alt.Text(field='min')),
          alt.Chart(data=pd.DataFrame({'max':['_']}))
            .mark_text()
            .encode(alt.Text(field='max'))
          
          )

      charts.append(spacer)
      charts[chart_index] = alt.vconcat(charts[chart_index],bounds, spacing=0)

  merged = alt.hconcat(*charts, spacing=chart_padding)# 8 seconds

  # for each column to view
  text_width = column_width - chart_padding
  table_width = column_width * (len(vars)+1)


  
  table = (alt.Chart(data)
    .transform_filter(brush)
    .transform_filter(picker)
    .transform_window(window=[{"op":"row_number","as":"row_num"}])
    .transform_filter(
        alt.FieldLTEPredicate(field='row_num', lte=row_limit)
    )
    .transform_fold(list(vars))
    .mark_text(align='center', limit=text_width)
    .encode(
        alt.Y(field="row_num",type="ordinal", axis=None),
        alt.Text(field="value",type="nominal"),
        alt.X(field="key",sort=list(vars), type="nominal",scale=alt.Scale(padding=chart_padding,rangeMax=table_width),axis=alt.Axis(orient="top",labelAngle= 0,title=None, domain=False, ticks=False)))
  ).properties(width=table_width).add_selection(brush).add_selection(picker)
  return alt.vconcat(merged,table,spacing=0)



import altair as alt
from .distributional import hist
from .relational import scatterplot
import numpy as np

def pairplot(data=None,variables=None):
  if data is None:
    raise ValueError('[pairgrid] data cannot be null')

  if variables is None:
    # get all numeric columns from the dataframe 
    variables = list(data.select_dtypes(include=[np.number]).columns.values)

  grid = []
  
  for row_index,row_variable in enumerate(variables):
    grid_row = []
    for column_index,column_variable in enumerate(variables):
      chart = None
      xAxis = None
      yAxis = None

      if row_index == len(variables)-1 :
         xAxis = alt.Axis()

      if column_index == 0:
         yAxis = alt.Axis()

      sel = alt.selection_interval(name='paintbrush',resolve="intersect",encodings=['x','y'])
      
      if row_variable == column_variable:
       
        interactive = alt.selection_interval(name='paintbrush',resolve="intersect",encodings=['x'])
        # ,
        chart = hist(data,filters=[sel],interactive=interactive,x=column_variable,yAxis=yAxis,xAxis=xAxis).properties(height=100,width=100)
        if row_index == 0:
          chart.layer[0].encoding.y.title = row_variable
        if row_index == len(variables)-1:
          chart.layer[0].encoding.x.title = row_variable
        
      else:
       
        chart = scatterplot(data,x=column_variable,y=row_variable,filters=[sel],yAxis=yAxis,xAxis=xAxis).add_selection(sel).properties(height=100,width=100)
      grid_row.append(chart)

    grid.append(grid_row)

  chart = None
  for row_index in range(len(grid)):
    row = None
    for column_index in range(len(grid[row_index])):
       if row is None:
          row = grid[row_index][column_index]
       else: 
          row = alt.hconcat(row,grid[row_index][column_index],spacing=5)
      

    if chart is None:
      chart = row
    else: 
      chart = alt.vconcat(chart,row,spacing=5)
  return chart